/*
 * Author: Radovan Cerveny
 */

#include "BackwardNondeterministicDAWGMatching.hpp"

#include <exception/AlibException.h>
#include <string/LinearString.h>
#include <alphabet/Symbol.h>

#include <map>
#include <measure>

namespace stringology {

namespace exact {

std::set < unsigned > BackwardNondeterministicDAWGMatching::match ( const string::String & subject, const string::String & pattern ) {
    return getInstance ( ).dispatch ( subject.getData ( ), pattern.getData ( ) );
}

std::set < unsigned > BackwardNondeterministicDAWGMatching::match ( const string::LinearString & subject, const string::LinearString & pattern ) {
    using Bitmask = unsigned long long int;

    std::set < unsigned > occ;

    measurements::start ( "Preprocess", measurements::Type::PREPROCESS );

    std::map < alphabet::Symbol, Bitmask > symbolBitmaskLookupTable;
    bool longPattern;
    size_t bitmaskLength;
    Bitmask highestBitBitmask;
    Bitmask allOnesBitmask;

     // Setup helper variables
    longPattern = 64 < pattern.getContent ( ).size ( );
    bitmaskLength = longPattern ? 64 : pattern.getContent ( ).size ( );

    highestBitBitmask = 1ULL << ( bitmaskLength - 1 );
    allOnesBitmask = ( highestBitBitmask << 1 ) - 1;

     // Initialize the bitmasks with zeros for each symbol in the alphabet
    for ( const auto & symbol : pattern.getAlphabet ( ) )
        symbolBitmaskLookupTable[symbol] = 0ULL;

     // Mark the position in the bitmask for each symbol in the pattern
    for ( size_t i = 0; i < bitmaskLength; i++ )
        symbolBitmaskLookupTable[pattern.getContent ( ).at ( i )] |= 1ULL << ( bitmaskLength - i - 1 );

    measurements::end ( );

    measurements::start ( "Algorithm", measurements::Type::ALGORITHM );

    size_t posInSubject = 0;

    while ( posInSubject <= subject.getContent ( ).size ( ) - pattern.getContent ( ).size ( ) ) {
        size_t posInPattern = bitmaskLength;
        size_t lastPosOfFactor = bitmaskLength;
        Bitmask currentBitmask = ~0ULL;

        while ( currentBitmask != 0ULL ) {
            currentBitmask = currentBitmask & symbolBitmaskLookupTable[subject.getContent ( ).at ( posInSubject + posInPattern - 1 )];
            posInPattern--;

            if ( ( currentBitmask & highestBitBitmask ) != 0 ) {
                if ( posInPattern > 0 ) {
                    lastPosOfFactor = posInPattern;
                } else {
                    if ( !longPattern ) {
                         // Yay, there is match!!!
                        occ.insert ( posInSubject );
                    } else {
                         // if the pattern is longer then 64 characters switch to brute force check
                        size_t k = bitmaskLength;

                        while ( k < pattern.getContent ( ).size ( ) && pattern.getContent ( ).at ( k ) == subject.getContent ( ).at ( posInSubject + k ) ) k++;

                        if ( k == pattern.getContent ( ).size ( ) )
                             // Yay, there is match!!!
                            occ.insert ( posInSubject );
                    }
                }
            }

            currentBitmask <<= 1;

             // We need to trim excess ones in case the pattern is shorter then 64 characters
            currentBitmask &= allOnesBitmask;
        }

        posInSubject += lastPosOfFactor;
    }

    measurements::end ( );

    return occ;
}

auto BackwardNondeterministicDAWGMatchingLinearStringLinearString = BackwardNondeterministicDAWGMatching::RegistratorWrapper < std::set < unsigned >, string::LinearString, string::LinearString > ( BackwardNondeterministicDAWGMatching::getInstance ( ), BackwardNondeterministicDAWGMatching::match );

} /* namespace exact */

} /* namespace stringology */
